/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _LEVELDATA_H_INCLUDED_
#define _LEVELDATA_H_INCLUDED_

namespace Sith
{
	namespace World
	{
		// LevelData class
		// - Stores values that describe the entire level
		class LevelData
		{
		public:
			// Version of the level file.
			// Valid values: 1, for a standard Jedi Knight level.
			int Version;

			// Downward acceleration in JKUs.
			float Gravity;

			// Ceiling sky values
			float CeilingSkyZ;						// Height of the sky
			Math::Vector CeilingSkyOffset;	// U/V offset of the sky, used by Animation

			// Horizon sky values
			float HorizonDistance;					// Distance of the sky texture (ignored)
			float HorizonPixelsPerRev;				// Rotational velocity of the sky (ignored)
			Math::Vector HorizonSkyOffset;	// U/V offset, used by Animation

			// LOD values
			float MipMapDistances[4];				// Ignored - legacy data
			float LODDistances[4];					// Ignored - legacy data
			float PerspectiveDistance;				// Ignored - legacy data
			float GouraudDistance;					// Ignored - legacy data

            // Extensions
			UI::UIColor LevelTint[3];				// R/G/B Tint of the level, used by Animation
		};

		// LevelVertex class
		// - Stores informatino about a single vertex
		class LevelVertex
		{
		public:
			Math::Vector Location;	// Position of the vertex in the world
			Math::Vector Texture;	// Texture coordinates, X=U, Y=V, Z=W=0.
			Math::Vector Normal;	// Uninitialized...?
			UI::UIColor Color;				// Color of vertex (incl. lighting)
		};

		class LevelSector;

		// LevelSurface class
		// - Stores information about a single surface
		class LevelSurface
		{
		public:
			enum SurfaceFlag
			{
				SF_FLOOR = 0x01,
				SF_COG_LINKED = 0x02,
				SF_PLAYER_NOWALK = 0x04,
				SF_AI_NOWALK = 0x08,
				SF_UNKNOWN1 = 0x10,
				SF_UNKNOWN2 = 0x20,
				SF_UNKNOWN3 = 0x40,
				SF_NO_FALLING_DAMAGE = 0x80,
				SF_UNKNOWN4 = 0x100,
				SF_HORIZON_SKY = 0x200,
				SF_CEILING_SKY = 0x400,
				SF_SCROLLING = 0x800,
				SF_NO_DRAG = 0x1000,
				SF_SPEEDY = 0x2000,
				SF_MAGSEALED = 0x4000,
				SF_UNKNOWN5 = 0x8000,
				
				SF_METALFLOOR = 0x10000,
				SF_WATERFLOOR = 0x20000,
				SF_DAMPFLOOR = 0x40000,
				SF_DIRTFLOOR = 0x80000,
				SF_SOGGYFLOOR = 0x100000
			};

			enum FaceFlag
			{
				FF_UNKNOWN1 = 0x01,
				FF_TRANSLUCENT = 0x02,
				FF_UNKNOWN2 = 0x04
			};

			enum AdjoinFlag
			{
				AF_RENDER_THROUGH_ADJOIN = 0x01,
				AF_IGNORE_COLLISION = 0x02,
				AF_UNKNOWN1 = 0x04,
				AF_UNKNOWN2 = 0x08,
				AF_FORCE_PLAYER_COLLISION = 0x10,
				AF_COG_DISABLED_RENDERING = 0x20,
				AF_UNKNOWN3 = 0x40,
				AF_UNKNOWN4 = 0x80,
			};

			enum GEOMODE
			{
				GEO_NOT_RENDERED = 0,
				GEO_VERTEX = 1,
				GEO_WIREFRAME = 2,
				GEO_SOLIDCOLOR = 3,
				GEO_TEXTURED = 4
			} GeoMode;

			enum LIGHTMODE
			{
				LM_FULLBRIGHT = 0,
				LM_NOLIGHT = 1,
				LM_DIFFUSE = 2,
				LM_GOURAUD = 3
			} LightMode;

			enum TEXMODE
			{
				TM_AFFINE = 0,
				TM_PERSPECTIVE = 1
			} TextureMode;

			unsigned int SurfaceNumber;		// Number of the surface in the level
			LevelSector* ParentSector;

			unsigned int SurfaceFlags;
			unsigned int FaceFlags;

			// Texturing information
			Sith::VFS::WisePointer<Res::Material> Texture;
			int CurrentTextureCel;
			Math::Vector TextureShift;

			// Adjoin information:
			bool isAdjoined;
			unsigned int AdjoinFlags;
			LevelSurface* MirrorSurface;
			LevelSector* AdjoinedSector;

			// Geometry information
			std::vector<LevelVertex> VertexList;
		};

		// LevelSector class
		// - Stores information about a single sector; precompiled surface lists, etc.
		class LevelSector
		{
		public:
			enum SectorFlag
			{
				SF_IGNORE_GRAVITY = 0x01,
				SF_WATER = 0x02,
				SF_COG_LINKED = 0x04,
				SF_USE_THRUST = 0x08,
				SF_NOT_ON_MAP = 0x10,
				SF_UNKNOWN1 = 0x20,
				SF_BOTTOMLESS_PIT = 0x40,
				SF_ADJOINS_DISABLED = 0x80,
				SF_UNKNOWN2 = 0x100,
				SF_UNKNOWN3 = 0x200,
				SF_UNKNOWN4 = 0x400,
				SF_UNKNOWN5 = 0x800,
				SF_COLLIDEBOX = 0x1000,
				SF_UNKNOWN6 = 0x2000,
				SF_ALWAYS_ON_MAP = 0x4000
			};

			unsigned int SectorNumber;		// Number of the sector in the level
			unsigned int SectorFlags;		// Sector flags - see above enumerator

			// Sector geometry
			std::vector<LevelSurface> Geometry;

			// Sector properties (physics)
			Math::Vector Center;	// Center of the sector
			float Radius;					// Radius of the sector
			float BoundBox[6];				// Temporary; will probably be removed
			float CollideBox[6];			// Temporary

			// Graphical effects:
			float AmbientLight;				// Ambient lighting inside the sector (shows up on internal models, IIRC)
			float ExtraLight;				// Extra lighting on the geo itself
			UI::UIColor Tint;				// Sector tint - set on screen when camera enters it
		};

		class LevelLight
		{
		public:
			float X, Y, Z;
			float Intensity;
			float Range;
		};
	};
};

#endif // _LEVELDATA_H_INCLUDED_