/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "vmtype.h"
#include "opcodes.h"
#include "instruction.h"
#include "vmheap.h"
#include "stackmachine.h"
#include "coglib.h"

// Temporary StackMachine constructor
// - Instantiates variables
// - Loads the COG script
// - Sends the startup message
Sith::COG::StackMachine::StackMachine()
{
	InstructionCount = 0;
	PulseAccumulator = 0;
	SleepAccumulator = 0;
	Pulse = 0;
	SleepTime = 0;
	return;
};

Sith::COG::StackMachine::~StackMachine()
{
	return;
};

// SendMessage()
// - Work in progress. For now, it doesn't do anything except jump into the COG script
// - Should add the Message to a message list or stack so they can be processed in order
// and not preempt other scripts or the gamestate object
Sith::COG::VMType Sith::COG::StackMachine::SendMessage(std::string MessageName, VMType Sender, int OverrideID, VMType Source, VMType Par1, VMType Par2, VMType Par3, VMType Par4)
{
	int ho = Heap.TranslateLocalHeapName(MessageName);

	// Early escape if this script does not carry a message entry by this name:
	if(ho == -1)
	{
		// Message not found
		return COG::VMType();
	};

	// This should troll the heap for relevant link information, unless the sender is VMTC_VOID in which case it should use the OverrideID.
	if(Sender.GetType() == VMTC_VOID)
	{
		return Execute(Heap.Get(ho, -1).GetAddr(), Sender, OverrideID, Source, Par1, Par2, Par3, Par4);
	}

	for(int i = 0; i < Heap.Local.HeapCount; ++i)
	{
		VMType Scratch = Heap.Get(i, -1);
		if((Scratch.GetType() == Sender.GetType()) && (Scratch.GetInt() == Sender.GetInt()))
		{
            return Execute(Heap.Get(ho, -1).GetAddr(), Sender, Heap.GetExtensions(i, -1).LinkID, Source, Par1, Par2, Par3, Par4);
		};
	};

	// No Link ID found
	return Execute(Heap.Get(ho, -1).GetAddr(), Sender, OverrideID, Source, Par1, Par2, Par4, Par4);
};

// Durr i wonder what this does
Sith::COG::VMType Sith::COG::StackMachine::Execute(int StartOffset, VMType Sender, int SenderID, VMType Source, VMType Par1, VMType Par2, VMType Par3, VMType Par4)
{
	// Set up return:
	COG::VMType RetVal = COG::VMType();

	// Check for a valid start offset
	if((StartOffset < 0) || (StartOffset > InstructionCount))
	{
		return RetVal;
	};
	
	// Loop through the instruction list
	for(int i = StartOffset; i < InstructionCount; ++i)
	{
		Sith::COG::Instruction& InstrRef = Code[i];
		switch(InstrRef.OPCODE)
		{
			// SYSCALL - Serves the function of a software or hardware interrupt
		case Opcodes::SYSCALL:
			switch(InstrRef.DATA.GetAddr())
			{
			case 0:
				// GET SENDER REF
				ReturnRegister = Sender.GetInt();
				break;

			case 1:
				// GET SOURCE REF
				ReturnRegister = Source.GetInt();
				break;

			case 2:
				// GET SENDER ID
				ReturnRegister = SenderID;
				break;

			case 3:
				// SLEEP
				ReturnRegister = VMType();
				SleepJump = i + 1;
				SleepSource = Source;
				SleepSender = Sender;
				SleepSenderID = SenderID;
				SleepPar1 = Par1;
				SleepPar2 = Par2;
				SleepPar3 = Par3;
				SleepPar4 = Par4;
				SleepTime = Stack.top().GetFloat();
				Stack.pop();
				SleepAccumulator = 0.0;
				return RetVal;

			case 4:
				// GETPARAM(n)
				{
					int ParamNum = Stack.top().GetInt();
					Stack.pop();
					if(ParamNum == 0) ReturnRegister = Par1;
					else if(ParamNum == 1) ReturnRegister = Par2;
					else if(ParamNum == 2) ReturnRegister = Par3;
					else if(ParamNum == 3) ReturnRegister = Par4;
					else ReturnRegister = VMType();
				}
				break;

			case 5:
				// RETURNEX(n)
				{
					ReturnRegister = VMType();
					RetVal = Stack.top();
					Stack.pop();
				};
				break;

			default:
				break;
			};
			break;
		
			// LIBCALL - Calls functions in the attached COGLIB
		case Opcodes::LIBCALL:
			{
				int Offset = InstrRef.DATA.GetAddr();
				CogLib::Call(Offset, this);
			}
			break;
			
			// MSGCALL - Calls a local message (short jump)
			// To call external messages, use the COGLIB function SendMessage[Ex]() or somesuch.
		case Opcodes::MSGCALL:
			{
				int Addr = 0;
				Addr = Heap.Get(InstrRef.DATA.GetAddr(), InstrRef.DATA.GetAddrBoost()).GetAddr();
				Execute(Addr, VMType(), -1, VMType(), VMType(), VMType(), VMType(), VMType());
			};
			break;		
		
			// RETURN - Returns control to the calling function
			// Note: The stack is preserved, so this will safely pass return values to the caller
			// assuming both this and the caller are running on the same stack machine or execution instance
		case Opcodes::RETURN:
			return RetVal;
		
		case Opcodes::LOAD:	// LOAD - Pushes the heap value at the topmost address onto the stack
			Stack.push(Heap.Get(InstrRef.DATA.GetAddr(), InstrRef.DATA.GetAddrBoost()));
			break;
		
		case Opcodes::STOR:	// STOR - Stores the second-topmost value at the topmost address
			Heap.Set(InstrRef.DATA.GetAddr(), InstrRef.DATA.GetAddrBoost(), Stack.top());
			Stack.pop();
			break;
			
		case Opcodes::ILOD:	// ILOD - Pushes the heap value at the address + arbitrary index onto the stack
			{
				int index = InstrRef.DATA.GetAddr() + Stack.top().GetInt();
				Stack.pop();
				Stack.push(Heap.Get(index, InstrRef.DATA.GetAddrBoost()));
			};
			break;
			
			// ISTR - Stores the VMType on the top of the stack in the heap, at Heap[DATA + second_topmost]
		case Opcodes::ISTR:
			{
				VMType Var = Stack.top();
				Stack.pop();
				int index = InstrRef.DATA.GetAddr() + Stack.top().GetInt();
				Stack.pop();
				Heap.Set(index, InstrRef.DATA.GetAddrBoost(), Var);
			};
			break;
		
			// PUSH - pushes a value onto the stack
		case Opcodes::PUSH:
			Stack.push(VMType(InstrRef.DATA));
			break;
		
			// POP - pops a value off of the stack
		case Opcodes::POP:
			Stack.pop();
			break;
		
			// FLSH - flushes the stack
			// This is used to clean up after the 'dirty stack' model of code generation
			// You'll usually find this near every return and after every expression statement
		case Opcodes::FLSH:
			while(Stack.empty() == false)
			{
				Stack.pop();
			};
			break;

			// RPSH - pushes the return register onto the stack
		case Opcodes::RPSH:
			Stack.push(VMType(ReturnRegister));
			break;

		case Opcodes::ADD:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue + RValue));
			};
			break;
		
		case Opcodes::SUB:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue - RValue));
			};
			break;
		
		case Opcodes::MUL:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue * RValue));
			};
			break;
		
		case Opcodes::DIV:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue / RValue));
			};
			break;
		
		case Opcodes::MOD:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue % RValue));
			};
			break;
		
		case Opcodes::AND:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue & RValue));
			};
			break;
		
		case Opcodes::OR:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue | RValue));
			};
			break;
		
		case Opcodes::XOR:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(LValue ^ RValue));
			};
			break;
			
		case Opcodes::NOT:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(bool(!Value.GetBool())));
			};
			break;
		
		case Opcodes::INC:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(++Value));
			};
			break;
		
		case Opcodes::DEC:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(--Value));
			};
			break;
		
		case Opcodes::SHL:
			{
				int Arg = Stack.top().GetInt();
				Stack.pop();
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(Value << Arg));
			};
			break;
		
		case Opcodes::SHR:
			{
				int Arg = Stack.top().GetInt();
				Stack.pop();
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				Stack.push(VMType(Value >> Arg));
			};
			break;
		
		case Opcodes::CZ:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				if(Value == VMType(0))
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CNZ:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				if(Value != VMType(0))
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue == RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CNE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue != RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CGT:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue > RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CLT:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue < RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CGTE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue >= RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::CLTE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue <= RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
			
		case Opcodes::CT:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				if(Value == true)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
			
		case Opcodes::CNT:
			{
				VMType Value;
				Value = Stack.top();
				Stack.pop();
				
				if(Value == false)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
			
		case Opcodes::CAND:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue && RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
			
		case Opcodes::COR:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue || RValue)
				{
					Stack.push(VMType(true));
				}
				else
				{
					Stack.push(VMType(false));
				};
			};
			break;
		
		case Opcodes::JMP:
			i = InstrRef.DATA.GetAddr() - 1;
			break;
		
		case Opcodes::JZ:
			{
				VMType Value = Stack.top();
				Stack.pop();
				
				if(Value == 0)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JNZ:
			{
				VMType Value = Stack.top();
				Stack.pop();
				
				if(Value != 0)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue == RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JNE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue != RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JGT:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue > RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JLT:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue < RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JGTE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue >= RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
		
		case Opcodes::JLTE:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue <= RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
			
		case Opcodes::JT:
			{
				VMType Value = Stack.top();
				Stack.pop();
				
				if(Value == true)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
			
		case Opcodes::JNT:
			{
				VMType Value = Stack.top();
				Stack.pop();
				
				if(Value == false)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
			
		case Opcodes::JAND:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue && RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
			
		case Opcodes::JOR:
			{
				VMType LValue, RValue;
				RValue = Stack.top();
				Stack.pop();
				LValue = Stack.top();
				Stack.pop();
				
				if(LValue || RValue)
				{
					i = InstrRef.DATA.GetAddr() - 1;
				};
			};
			break;
			
			// :)
		case Opcodes::ZEQ:
			break;
		
		case Opcodes::NOP:
		default:
			break;
		};
	};
	return RetVal;
};
