/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _VECTOR_H_INCLUDED_
#define _VECTOR_H_INCLUDED_

namespace Sith
{
	namespace Math
	{
		class Vector
		{
		public:
			float x, y, z;

			Vector()
			{
			}

			Vector(const float _x, const float _y, const float _z = 0.0f)
			{
				x = _x;	y = _y;	z = _z;
			}

			Vector(const float *p)
			{
				x = p[0]; y = p[1]; z = p[2];
			}

			Vector(const Vector &a, const Vector &b)
			{
				x = a.x - b.x;
				y = a.y - b.y;
				z = a.z - b.z;
			}


			Vector operator + (const Vector &a)
			{
				return Vector(x + a.x, y + a.y, z + a.z);
			}
			void operator += (const Vector &a)
			{
				x += a.x;
				y += a.y;
				z += a.z;
			}
			Vector operator + (const float a)
			{
				return Vector(x + a, y + a, z + a);
			}
			void operator += (const float a)
			{
				x += a;
				y += a;
				z += a;
			}


			Vector operator - (const Vector &a)
			{
				return Vector(x - a.x, y - a.y, z - a.z);
			}
			void operator -= (const Vector &a)
			{
				x -= a.x;
				y -= a.y;
				z -= a.z;
			}
			Vector operator - (const float a)
			{
				return Vector(x - a, y - a, z - a);
			}
			void operator -= (const float a)
			{
				x -= a;
				y -= a;
				z -= a;
			}


			Vector operator * (const Vector &a)
			{
				return Vector(x * a.x, y * a.y, z * a.z);
			}
			void operator *= (const Vector &a)
			{
				x *= a.x;
				y *= a.y;
				z *= a.z;
			}
			Vector operator * (const float a)
			{
				return Vector(x * a, y * a, z * a);
			}
			void operator *= (const float a)
			{
				x *= a;
				y *= a;
				z *= a;
			}


			Vector operator / (const Vector &a)
			{
				return Vector(x / a.x, y / a.y, z / a.z);
			}
			void operator /= (const Vector &a)
			{
				x /= a.x;
				y /= a.y;
				z /= a.z;
			}
			Vector operator / (const float a)
			{
				float ftemp = 1 / a;
				return Vector(x * ftemp, y * ftemp, z * ftemp);
			}
			void operator /= (const float a)
			{
				float ftemp = 1 / a;
				x *= ftemp;
				y *= ftemp;
				z *= ftemp;
			}


			bool operator == (Vector param)
			{
				if (x == param.x && y == param.y && z == param.z)
					return 1;
				return 0;
			}
			bool operator != (Vector param)
			{
				if (x != param.x || y != param.y ||z != param.z)
					return 1;
				return 0;
			}


			Vector& Normalize()
			{
				float invlen = 1 / sqrt(x * x + y * y + z * z);
				x *= invlen;
				y *= invlen;
				z *= invlen;
				return *this;
			}
			// normalizes, but not if zero length
			Vector& ZeroNormalize()
			{
				float invlen = sqrt(x * x + y * y + z * z);
				if (invlen == 0)
					return *this;
				invlen = 1 / invlen;
				x *= invlen;
				y *= invlen;
				z *= invlen;
				return *this;
			}

			void Clamp(const float lower, const float upper)
			{
				if (x > upper) x = upper; else if (x < lower) x = lower;
				if (y > upper) y = upper; else if (y < lower) y = lower;
				if (z > upper) z = upper; else if (z < lower) z = lower;
			}

			float Length()
			{
				return sqrt(x * x + y * y + z * z);
			}
			float LengthSq()
			{
				return x * x + y * y + z * z;
			}

			void Set(const float a, const float b, const float c)
			{
				x = a; y = b; z = c;
			}
		};


		inline Vector operator * (const float param, const Vector &v)
		{
			return Vector(param * v.x, param * v.y, param * v.z);
		}


		inline Vector VectorAdd(const Vector &Vec1, const Vector &Vec2)
		{
			return Vector(Vec1.x + Vec2.x, Vec1.y + Vec2.y, Vec1.z + Vec2.z);
		};

		inline Vector VectorCross(const Vector &Vec1, const Vector &Vec2)
		{
			return Vector(Vec1.y * Vec2.z - Vec2.y * Vec1.z, Vec2.x * Vec1.z - Vec1.x * Vec2.z, Vec1.x * Vec2.y - Vec2.x * Vec1.y);
		};

		inline float VectorDistSq(const Vector &Vec1, const Vector &Vec2)
		{
			float xpart = Vec1.x - Vec2.x;
			float ypart = Vec1.y - Vec2.y;
			float zpart = Vec1.z - Vec2.z;

			return xpart * xpart + ypart * ypart + zpart * zpart;
		};

		inline float VectorDist(const Vector &Vec1, const Vector &Vec2)
		{
			float xpart = Vec1.x - Vec2.x;
			float ypart = Vec1.y - Vec2.y;
			float zpart = Vec1.z - Vec2.z;

			return sqrt(xpart * xpart + ypart * ypart + zpart * zpart);
		};

		inline float VectorDot(const Vector &Vec1, const Vector &Vec2)
		{
			return (Vec1.x * Vec2.x) + (Vec1.y * Vec2.y) + (Vec1.z * Vec2.z);
		};

		inline float VectorLenSq(const Vector &Vec1)
		{
			return (Vec1.x * Vec1.x) + (Vec1.y * Vec1.y) + (Vec1.z * Vec1.z);
		};

		inline float VectorLen(const Vector &Vec1)
		{
			return sqrt((Vec1.x * Vec1.x) + (Vec1.y * Vec1.y) + (Vec1.z * Vec1.z));
		};

		inline Vector VectorNorm(const Vector &Vec1)
		{
			float invlen = 1 / sqrt(Vec1.x * Vec1.x + Vec1.y * Vec1.y + Vec1.z * Vec1.z);
			return Vector(Vec1.x * invlen, Vec1.y * invlen, Vec1.z * invlen);
		};

		// normalizes, but not if zero length
		inline Vector VectorZeroNorm(const Vector &Vec1)
		{
			float invlen = sqrt(Vec1.x * Vec1.x + Vec1.y * Vec1.y + Vec1.z * Vec1.z);
			if (invlen == 0)
				return Vector(0, 0, 0);
			invlen = 1 / invlen;
			return Vector(Vec1.x * invlen, Vec1.y * invlen, Vec1.z * invlen);
		};

		inline Vector VectorScale(const Vector &Vec1, const float Scale)
		{
			return Vector(Vec1.x * Scale, Vec1.y * Scale, Vec1.z * Scale);
		};

		inline Vector VectorSet(const float x, const float y, const float z)
		{
			return Vector(x, y, z);
		};

		inline Vector VectorSub(const Vector &Vec1, const Vector &Vec2)
		{
			return Vector(Vec1.x - Vec2.x, Vec1.y - Vec2.y, Vec1.z - Vec2.z);
		};

		inline float VectorX(const Vector &Vec1)
		{
			return Vec1.x;
		};

		inline float VectorY(const Vector &Vec1)
		{
			return Vec1.y;
		};

		inline float VectorZ(const Vector &Vec1)
		{
			return Vec1.z;
		};
	};
};

#endif // _VECTOR_H_INCLUDED_
