/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _VFS_H_INCLUDED_
#define _VFS_H_INCLUDED_

namespace Sith
{
	namespace VFS
	{
		class VirtualFileSystem
		{
		private:
			VDirectory ResourceRoot;
			VDirectory PathRoot;

			std::list<Sith::VFS::Episode> EpisodeList;
			Episode* CurrentEpisode;

			std::string EpisodePath;
			std::string ResourcePath;
			std::string ModPath;

			inline SITHERROR ParseEpisodeJK(FILE* EpisodeFile);
			inline SITHERROR ParseCOGStringsUNI(FILE* COGStringsFile);

			inline const char* OpenFSBuffer(FILE* FileHandle)
			{
				char* Buffa = 0;
				long int FSLength = 0;

				fseek(FileHandle, 0L, SEEK_END);
				FSLength = ftell(FileHandle);

				Buffa = new char[FSLength + 1];
				if(!Buffa)
				{
					ConOut("ERROR: Out of memory, failed to allocate buffer for file system operation.\n", ConOut::Error);
					return 0;
				};

				Buffa[FSLength] = '\0';

				fseek(FileHandle, 0L, SEEK_SET);

				size_t BytesRead = fread(Buffa, sizeof(char), FSLength, FileHandle);
				if(BytesRead != FSLength)
				{
					ConOut("ERROR: File size indicated by operating system does not match physical data size.\n", ConOut::Error);
					free(Buffa);
					return 0;
				};

				fclose(FileHandle);
				return Buffa;
			};

			VirtualFileSystem();

		public:
			static inline VirtualFileSystem& Get()
			{
				static VirtualFileSystem vfs;
				return vfs;
			};

			SITHERROR InitializeResource(std::string theResourcePath);
			SITHERROR InitializePath(std::string theModPath);
			SITHERROR InitializeEpisode(std::string theEpisodePath);
			SITHERROR Shutdown();

			FILE* Open(std::string FilePath, bool isAscii);
			const char* OpenBuffer(std::string FilePath);

			SITHERROR MountEpisode(std::string ShortName);
			SITHERROR UnmountEpisode();

			inline void ls()
			{
				ConOut("Scanning through Virtual File System...\n", ConOut::Standard);

				int Result = 0;

				Result += Sith::VFS::VirtualFileSystem::Get().ResourceRoot.ls("");
				Result += Sith::VFS::VirtualFileSystem::Get().PathRoot.ls("");
				if(Sith::VFS::VirtualFileSystem::Get().CurrentEpisode) Result += Sith::VFS::VirtualFileSystem::Get().CurrentEpisode->EpisodeContainer.ls("");

				std::ostringstream Stream;
				Stream << "Found " << Result << " files.\n\n";

				ConOut(Stream.str(), ConOut::Standard);
			};

			inline void lsepisode()
			{
				std::stringstream Stream;
				Stream << "The Virtual File System contains " << (unsigned int)Sith::VFS::VirtualFileSystem::Get().EpisodeList.size() << " episodes.\n";
				ConOut(Stream.str(), ConOut::Standard);

				for(std::list<Episode>::iterator it = EpisodeList.begin(); it != EpisodeList.end(); ++it)
				{
					Stream.str("");
					Stream << "\t| " << it->ShortName << ": \"" << it->Name << "\" (" << (unsigned int)it->LevelList.size() << ") - ";
					switch(it->Type)
					{
					case Episode::SINGLEPLAYER:
						Stream << "Singleplayer\n";
						break;
					case Episode::DEATHMATCH:
						Stream << "Deathmatch\n";
						break;
					case Episode::CAPTURETHEFLAG:
						Stream << "Capture the Flag\n";
						break;
					default:
						Stream << "Unknown Type\n";
						break;
					};

					ConOut(Stream.str(), ConOut::Standard);
				};

				ConOut("End of list\n\n", ConOut::Standard);

			};

			inline void lsmap()
			{
				if(!CurrentEpisode)
				{
					ConOut("No episode is currently mounted; there are no levels visible to the Virtual File System.\n", ConOut::Standard);
				}
				else
				{
                    std::stringstream Stream;
					Stream << "The current episode, \'" << CurrentEpisode->ShortName << "\', contains " << (unsigned int)CurrentEpisode->LevelList.size() << " ";
					switch(CurrentEpisode->Type)
					{
					case Episode::SINGLEPLAYER:
						Stream << "Singleplayer";
						break;
					case Episode::DEATHMATCH:
						Stream << "Deathmatch";
						break;
					case Episode::CAPTURETHEFLAG:
						Stream << "Capture the Flag";
						break;
					default:
						Stream << "Unknown";
						break;
					};
					Stream << " levels.\n";

					ConOut(Stream.str(), ConOut::Standard);

					for(std::list<Episode::EpisodeElement>::iterator it = CurrentEpisode->LevelList.begin(); it != CurrentEpisode->LevelList.end(); ++it)
					{
						Stream.str("");
						Stream << "\t| " << Sith::COG::CogLib::GetUniString(it->ShortName) << " (" << it->ShortName << ")\n";

						ConOut(Stream.str(), ConOut::Standard);
					};

					ConOut("End of list\n\n", ConOut::Standard);
				};

				return;
			};

			inline void mapinfo(std::string ShortName)
			{
				if(!CurrentEpisode)
				{
					ConOut("No episode is currently mounted; there are no levels visible to the Virtual File System.\n", ConOut::Standard);
					return;
				};

				for(std::list<Episode::EpisodeElement>::iterator it = CurrentEpisode->LevelList.begin(); it != CurrentEpisode->LevelList.end(); ++it)
				{
					if(it->ShortName == ShortName)
					{
						std::ostringstream Stream;
						Stream << "Detailed map information:\n";
						Stream << "\t" << Sith::COG::CogLib::GetUniString(it->ShortName) << "\n\t\t";
						Stream << Sith::COG::CogLib::GetUniString(it->ShortName + "_TEXT_00") << "\n\t\t";
						Stream << Sith::COG::CogLib::GetUniString(it->ShortName + "_TEXT_01") << "\n";

						ConOut(Stream.str(), ConOut::Standard);

						return;
					};
				};

				ConOut("Could not find level \'" + ShortName + "\' in the currently mounted episode.\n", ConOut::Standard);

				return;
			};

			inline const std::list<Sith::VFS::Episode>& GetEpisodeList() const
			{
				return EpisodeList;
			};

			inline Sith::VFS::Episode*  & GetCurrentEpisode() 
			{
				return CurrentEpisode;
			};
		};
	};
};

#endif // _VFS_H_INCLUDED_