/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _VMHEAP_H_INCLUDED_
#define _VMHEAP_H_INCLUDED_

namespace Sith
{
	namespace COG
	{
		// VMHeap class
		// - Stores the heap, names, desc, any additional extensions that happen along
		// - Obviously this isn't done yet.
		class VMHeap
		{
		public:
			struct SymbolExtensionTable
			{
				bool isLocal;
				bool NoLink;

				std::string Desc;

				int Mask;
				int LinkID;

				SymbolExtensionTable()
				{
					isLocal = false;
					NoLink = false;

					Desc = "";

					Mask = 0;
					LinkID = -1;

					return;
				};
			};

		private:
			class HeapElement
			{
			public:
				std::string Name;
				VMType Data;
				int Address;

				SymbolExtensionTable Extensions;

				HeapElement(std::string _Name, VMType _Data, int _Address) : Name(_Name), Data(_Data), Address(_Address)
				{
					return;
				};
			};

		public:
			class Namespace
			{
			public:
				std::string Name;
				int Address;

				std::vector<HeapElement> Heap;
				int HeapCount;

				Namespace(std::string _Name, int _Address) : Name(_Name), Address(_Address), HeapCount(0)
				{
					return;
				};
			};

			Namespace Local;

		private:
			static std::vector<Namespace> Global;
			static int GlobalCount;

		public:
			VMHeap() : Local("local", -1)
			{
				return;
			};

			inline SITHERROR AddLocal(std::string Name, VMType Default)
			{
				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<HeapElement>::iterator it = Local.Heap.begin(); it != Local.Heap.end(); ++it)
				{
					if(it->Name == Name)
					{
						std::ostringstream Stream;
						Stream << "WARNING: Tried to re-create local symbol \'" << Name << "\'.\n";
						ConOut(Stream.str(), ConOut::Warning);
						return SE_COG_WARNING_ALREADY_EXISTS;
					};
				};

				Local.Heap.push_back(HeapElement(Name, Default, Local.HeapCount++));
				return SE_OK;
			};

			inline SITHERROR AddLocal(std::string Name, VMType Default, SymbolExtensionTable Ext)
			{
				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<HeapElement>::iterator it = Local.Heap.begin(); it != Local.Heap.end(); ++it)
				{
					if(it->Name == Name)
					{
						std::ostringstream Stream;
						Stream << "WARNING: Tried to re-create local symbol \'" << Name << "\'.\n";
						ConOut(Stream.str(), ConOut::Warning);
						return SE_COG_WARNING_ALREADY_EXISTS;
					};
				};

				Local.Heap.push_back(HeapElement(Name, Default, Local.HeapCount++));
				Local.Heap.back().Extensions = Ext;
				return SE_OK;
			};

			static inline SITHERROR AddNamespace(std::string Name)
			{
				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<Namespace>::iterator it = Global.begin(); it != Global.end(); ++it)
				{
					if(it->Name == Name)
					{
						std::ostringstream Stream;
						Stream << "WARNING: Tried to re-create namespace \'" << Name << "\'.\n";
						ConOut(Stream.str(), ConOut::Warning);
						return SE_COG_WARNING_ALREADY_EXISTS;
					};
				};

				Global.push_back(Namespace(Name, GlobalCount++));
				return SE_OK;
			};

			static inline int TranslateNamespaceName(std::string Name)
			{
				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<Namespace>::iterator it = Global.begin(); it != Global.end(); ++it)
				{
					if(it->Name == Name)
					{
						return it->Address;
					};
				};

				// Doesn't exist, return local
				return -1;
			};

			inline int TranslateLocalHeapName(std::string Name)
			{
				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<HeapElement>::iterator it = Local.Heap.begin(); it != Local.Heap.end(); ++it)
				{
					if(it->Name == Name)
					{
						return it->Address;
					};
				};

				// Still here - couldn't find.
				// This is a *HACK* because LEC is god awful. Apparently you can prefix identifiers with 'cog' and they're still accepted.
				// Test for COG, strip COG, test again.
				if(Name.length() > 0)
				{
					if(strncmp("cog", Name.c_str(), 3) == 0)
					{
						return TranslateLocalHeapName(std::string(&Name.c_str()[3]));
					};
				};

				return -1;
			};

			static inline int TranslateNamespaceHeapName(int NSAddr, std::string Name)
			{
				if((NSAddr < 0) || (NSAddr > GlobalCount))
				{
					return -1;
				};

				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<HeapElement>::iterator it = Global[NSAddr].Heap.begin(); it != Global[NSAddr].Heap.end(); ++it)
				{
                    if(it->Name == Name)
					{
						return it->Address;
					};
				};

				// Doesn't exist >:(
				return -1;
			};

			static inline SITHERROR AddGlobal(int NSAddr, std::string Name, VMType Default)
			{
				if((NSAddr < 0) || (NSAddr > GlobalCount))
				{
					return SE_COG_WARNING;
				};

				std::transform(Name.begin(), Name.end(), Name.begin(), tolower);
				for(std::vector<HeapElement>::iterator it = Global[NSAddr].Heap.begin(); it != Global[NSAddr].Heap.end(); ++it)
				{
					if(it->Name == Name)
					{
						std::ostringstream Stream;
						Stream << "WARNING: Tried to re-create global \'" << Global[NSAddr].Name << "::" << Name << "\'.\n";
						ConOut(Stream.str(), ConOut::Warning);
						return SE_COG_WARNING_ALREADY_EXISTS;
					};
				};

				Global[NSAddr].Heap.push_back(HeapElement(Name, Default, Global[NSAddr].HeapCount++));
				return SE_OK;
			};

			static inline VMType GetGlobal(int VarAddr, int NSAddr)
			{
				if((NSAddr < 0) || (NSAddr >= GlobalCount))
				{
					return VMType(COG::VMTC_VOID);
				}
				else
				{
					if((VarAddr < 0) || (VarAddr >= Global[NSAddr].HeapCount))
					{
						return VMType(COG::VMTC_VOID);
					}
					else
					{
						return Global[NSAddr].Heap[VarAddr].Data;
					};
				};
			};

			inline VMType Get(int VarAddr, int NSAddr)
			{
				if((NSAddr < 0) || (NSAddr >= GlobalCount))
				{
					if((VarAddr < 0) || (VarAddr >= Local.HeapCount))
					{
						return VMType(COG::VMTC_VOID);
					}
					else
					{
						return Local.Heap[VarAddr].Data;
					};
				}
				else
				{
					if((VarAddr < 0) || (VarAddr >= Global[NSAddr].HeapCount))
					{
						return VMType(COG::VMTC_VOID);
					}
					else
					{
						return Global[NSAddr].Heap[VarAddr].Data;
					};
				};
			};

			inline SymbolExtensionTable GetExtensions(int VarAddr, int NSAddr)
			{
				if((NSAddr < 0) || (NSAddr >= GlobalCount))
				{
					if((VarAddr < 0) || (NSAddr >= GlobalCount))
					{
						return SymbolExtensionTable();
					}
					else
					{
						return Local.Heap[VarAddr].Extensions;
					};
				}
				else
				{
					if((VarAddr < 0) || (VarAddr >= Global[NSAddr].HeapCount))
					{
						return SymbolExtensionTable();
					}
					else
					{
						return Global[NSAddr].Heap[VarAddr].Extensions;
					};
				};
			};

			inline SITHERROR Set(int VarAddr, int NSAddr, VMType NewData)
			{
				if((NSAddr < 0) || (NSAddr >= GlobalCount))
				{
					if((VarAddr < 0) || (VarAddr >= Local.HeapCount))
					{
						return SE_COG_WARNING;
					}
					else
					{
						Local.Heap[VarAddr].Data = NewData;
						return SE_OK;
					};
				}
				else
				{
					if((VarAddr < 0) || (VarAddr >= Global[NSAddr].HeapCount))
					{
						return SE_COG_WARNING;
					}
					else
					{
						Global[NSAddr].Heap[VarAddr].Data = NewData;
						return SE_OK;
					};
				};
			};

			inline void Clear()
			{
				Local.Address = 0;
				Local.HeapCount = 0;
				Local.Heap.clear();

				return;
			};

			inline unsigned int GetConfigurableValueCount()
			{
				unsigned int VarCnt = 0;
				for(std::vector<HeapElement>::iterator it = Local.Heap.begin(); it != Local.Heap.end(); ++it)
				{
					if(it->Extensions.isLocal)
					{
						continue;
					}
					else if(it->Data.GetType() == VMTC_INSTRUCTION_ADDRESS)
					{
						continue;
					}
					else
					{
						++VarCnt;
					};
				};

				return VarCnt;
			};

			int GetConfigurableValue(unsigned int ValueNumber)
			{
				unsigned int VarCnt = 0;
				for(std::vector<HeapElement>::iterator it = Local.Heap.begin(); it != Local.Heap.end(); ++it)
				{
					if(it->Extensions.isLocal)
					{
						continue;
					}
					else if(it->Data.GetType() == VMTC_INSTRUCTION_ADDRESS)
					{
						continue;
					}
					else if(VarCnt == ValueNumber)
					{
						return it->Address;
					}
					else
					{
						++VarCnt;
					};
				};

				return -1;
			};

			static SITHERROR SerializeGlobal(std::string TargetFileName);

			static COG::VMType* __cdecl lsns()
			{
				std::ostringstream Stream;
				Stream << "The COG Library currently registers " << (unsigned int)GlobalCount << " specialized namespaces.\n";
				ConOut(Stream.str(), ConOut::Standard);

				for(std::vector<Namespace>::iterator it = Global.begin(); it != Global.end(); ++it)
				{
					Stream.str("");
					Stream << "\t| " << it->Name << "(" << it->HeapCount << ");\n";
					ConOut(Stream.str(), ConOut::Standard);
				};

				ConOut("End of list\n\n", ConOut::Standard);

				return new COG::VMType(COG::VMTC_VOID);
			};

			static COG::VMType* __cdecl lsvar(COG::VMType* InString)
			{
				std::ostringstream Stream;

				int NSAddr = COG::VMHeap::TranslateNamespaceName(InString->GetString());
				if(NSAddr == -1)
				{
					Stream << "ERROR: \'" << InString->GetString() << "\' is not a valid namespace name.\n";
					ConOut(Stream.str(), ConOut::Error);
					
					return new VMType(VMTC_VOID);
				};

				Stream << "The namespace \'" << InString->GetString() << "\' contains " << Global[NSAddr].HeapCount << " symbols.\n";
				ConOut(Stream.str(), ConOut::Standard);

				for(std::vector<HeapElement>::iterator it = Global[NSAddr].Heap.begin(); it != Global[NSAddr].Heap.end(); ++it)
				{
					Stream.str("");
					Stream << "\t| " << it->Name << "\n";
					ConOut(Stream.str(), ConOut::Standard);
				};

				ConOut("End of list\n\n", ConOut::Standard);

				return new COG::VMType(COG::VMTC_VOID);
			};
		};
	};
};

#endif // _VMHEAP_H_INCLUDED_